//	Kaffeecontroller 1.0
//
//	Copyright (C) 2009 Paul Wilhelm <paul@mosfetkiller.de>
//	http://mosfetkiller.de/?s=kaffeecontroller
//
//	Modul: main.c | Hauptprogramm

//	Interner RC-Oszillator 8 MHz, CKDIV8 Enabled
#define F_CPU	1000000UL

//	Includes
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/io.h>

//	Bool
#define FALSE 0
#define TRUE 1

//	Bitpopelei
#define	set_bit(var, bit)	((var) |= (1 << (bit)))
#define	clear_bit(var, bit)	((var) &= (unsigned)~(1 << (bit)))

//	Pinout

//	Nulldurchgangserkennung
#define ZERO_CROSSING_w		PORTA
#define ZERO_CROSSING_r		PINA
#define ZERO_CROSSING_pin	0
#define ZERO_CROSSING_ddr	DDRA
#define ZERO_CROSSING_adc	0

//	Taster "1 Tasse"
#define BUTTON_1_CUP_w		PORTB
#define BUTTON_1_CUP_r		PINB
#define BUTTON_1_CUP_pin	4
#define BUTTON_1_CUP_ddr	DDRB

//	Taster "2 Tassen"
#define BUTTON_2_CUP_w		PORTB
#define BUTTON_2_CUP_r		PINB
#define BUTTON_2_CUP_pin	5
#define BUTTON_2_CUP_ddr	DDRB

//	Taster "Power"
#define BUTTON_POWER_w		PORTB
#define BUTTON_POWER_r		PINB
#define BUTTON_POWER_pin	6
#define BUTTON_POWER_ddr	DDRB

//	LED rot
#define LED_RED_w			PORTA
#define LED_RED_r			PINA
#define LED_RED_pin			3
#define LED_RED_ddr			DDRA

//	LED grün
#define LED_GREEN_w			PORTA
#define LED_GREEN_r			PINA
#define LED_GREEN_pin		1
#define LED_GREEN_ddr		DDRA

//	LED blau
#define LED_BLUE_w			PORTA
#define LED_BLUE_r			PINA
#define LED_BLUE_pin		2
#define LED_BLUE_ddr		DDRA

//	Hallschalter
#define SENSOR_MAGNET_w		PORTA
#define SENSOR_MAGNET_r		PINA
#define SENSOR_MAGNET_pin	4
#define SENSOR_MAGNET_ddr	DDRA
#define SENSOR_MAGNET_adc	3
#define SENSOR_MAGNET_thres	40	//	ADC Schwellwert

//	NTC
#define SENSOR_TEMP_w		PORTA
#define SENSOR_TEMP_r		PINA
#define SENSOR_TEMP_pin		5
#define SENSOR_TEMP_ddr		DDRA
#define SENSOR_TEMP_adc		4

//	TRIAC Boiler
#define TRIAC_BOILER_w		PORTA
#define TRIAC_BOILER_r		PINA
#define TRIAC_BOILER_pin	6
#define TRIAC_BOILER_ddr	DDRA

//	TRIAC Pumpe
#define TRIAC_PUMP_w		PORTA
#define TRIAC_PUMP_r		PINA
#define TRIAC_PUMP_pin		7
#define TRIAC_PUMP_ddr		DDRA

//	Schwellwert Taster-Haltezeit
#define BUTTON_THRESHOLD	30	//	ms

//	Min. Wassertemperatur
#define OPERATING_TEMPERATURE	100	//	ADC Schwellwert

//	Prototypen
void init();
void power_off();

//	Variablen
volatile unsigned char power = TRUE;	//	Power-Flag
volatile unsigned int time_counter, user_time_counter = 0;	//	Globaler und universeller Zeitzähler (ms)
volatile unsigned int button_1_cup_counter = 0, button_2_cup_counter = 0, button_power_counter = 0;	//	Tastenzähler
volatile unsigned char led_red_on = FALSE, led_green_on = FALSE, led_blue_on = FALSE;	//	LED-Statusflags
volatile unsigned char led_red_blink = FALSE, led_green_blink = FALSE, led_blue_blink = FALSE;	//	Flags für Blinken der LEDs
volatile unsigned char led_red_blink_offset = FALSE, led_green_blink_offset = FALSE, led_blue_blink_offset = FALSE;	//	Blinken versetzen
volatile unsigned char water = FALSE, temperature = FALSE, make_coffee = 0, pump_time = 0;	//	Wasser-, Temperatur-, Kaffeewunsch-Flags und Pumpzeit

//	Main
int main(void)
{
	//	Initialisierung
	init();

	//	Nach dem Einstecken des Netzsteckers gleich abschalten
	power_off();

	//	Hauptschleife
	while (1)
	{

		//	Wasserstand prüfen
		ADMUX = SENSOR_MAGNET_adc | (1 << ADLAR);	//	ADLAR
		set_bit(ADCSR, ADSC);
		loop_until_bit_is_clear(ADCSR, ADSC);
		unsigned char sense = ADCH;
		if (sense >= SENSOR_MAGNET_thres) water = TRUE; else water = FALSE;

		//	Temperatur prüfen
		ADMUX = SENSOR_TEMP_adc | (1 << ADLAR);	//	ADLAR
		set_bit(ADCSR, ADSC);
		loop_until_bit_is_clear(ADCSR, ADSC);
		sense = ADCH;
		if (sense >= OPERATING_TEMPERATURE) temperature = TRUE; else temperature = FALSE;


		//	Benutzer drückt Taster "Power"
		if (button_power_counter >= BUTTON_THRESHOLD)
		{
			//	Boiler aus (falls er eingeschaltet war)
			set_bit(TRIAC_BOILER_w, TRIAC_BOILER_pin);

			//	Kaffeewunsch löschen
			make_coffee = 0;

			//	LED-Flags aus
			led_red_on = led_green_on = led_blue_on = FALSE;
			led_red_blink = led_green_blink = led_blue_blink = FALSE;

			//	Warten, bis der Taster losgelassen wurde
			while (button_power_counter > 0)
			{
				if (button_power_counter > BUTTON_THRESHOLD / 2) button_power_counter = BUTTON_THRESHOLD / 2;
			}

			//	Abschalten (Schlafmodus)
			power_off();

			//	Taster entprellen
			button_power_counter = BUTTON_THRESHOLD;	//	Künstlichen Tastendruck erzeugen
			while (button_power_counter > 0);
		}

		//	Benutzer drückt Taster "1 Tasse"
		if (button_1_cup_counter >= BUTTON_THRESHOLD)
		{
			if (water && temperature)
			//	Maschine bereit
			{
				button_1_cup_counter = 0;

				make_coffee = 1;
			}
			else
			//	Maschine nicht bereit
			{
				//	Kaffeewunsch "1 Tasse" togglen (Kaffeewunsch "2 Tassen" wird ggf. gelöscht)
				if (make_coffee == 1)
				{
					make_coffee = 0;

					led_blue_blink = FALSE;
					led_blue_blink_offset = FALSE;
				}
				else
				{
					make_coffee = 1;

					led_blue_blink = TRUE;
					led_blue_blink_offset = TRUE;
				}

				//	Warten, bis der Taster losgelassen wurde
				while (button_1_cup_counter > 0)
				{
					if (button_1_cup_counter > BUTTON_THRESHOLD) button_1_cup_counter = BUTTON_THRESHOLD;
				}
			}
		}


		//	Benutzer drückt Taster "2 Tassen"
		if (button_2_cup_counter >= BUTTON_THRESHOLD)
		{
			if (water && temperature)
			//	Maschine bereit
			{
				button_2_cup_counter = 0;

				make_coffee = 2;
			}
			else
			//	Maschine nicht bereit
			{
				//	Kaffeewunsch "2 Tassen" togglen (Kaffeewunsch "1 Tasse" wird ggf. gelöscht)
				if (make_coffee == 2)
				{
					make_coffee = 0;

					led_blue_blink = FALSE;
					led_blue_blink_offset = FALSE;
				}
				else
				{
					make_coffee = 2;

					led_blue_blink = TRUE;
					led_blue_blink_offset = TRUE;
				}

				//	Warten, bis der Taster losgelassen wurde
				while (button_2_cup_counter > 0)
				{
					if (button_2_cup_counter > BUTTON_THRESHOLD) button_2_cup_counter = BUTTON_THRESHOLD;
				}
			}
		}


		//	Wasserstand prüfen
		if (water)
		//	Wasserstand ok -> Temperatur prüfen
		{
			led_red_blink = FALSE;

			if (temperature)
			//	Temperatur erreicht -> Auf Kaffeewunsch prüfen
			{
				//	Boiler aus (falls er eingeschaltet war)
				set_bit(TRIAC_BOILER_w, TRIAC_BOILER_pin);

				led_green_blink = FALSE;
				led_green_on = TRUE;

				if (make_coffee > 0)
				//	Kaffeewunsch vorhanden
				{
					led_blue_blink = FALSE;
					led_blue_blink_offset = FALSE;

					if (make_coffee == 1)
					//	1 Tasse Kaffee herstellen
					{
						//	Pumpzeit festlegen
						pump_time = 18;	//	s
					}
					else
					//	2 Tassen Kaffee herstellen
					{
						//	Pumpzeit festlegen
						pump_time = 36;	//	s
					}

					//	Zeitschleife
					user_time_counter = 0;
					unsigned char escape = FALSE;	//	Abbruchflag
					while (user_time_counter < (pump_time * 1000) && water && !escape)
					{
						//	Nulldurchgang (um festes Offset verschoben) erkennen
						ADMUX = ZERO_CROSSING_adc;
						set_bit(ADCSR, ADSC);
						loop_until_bit_is_clear(ADCSR, ADSC);
						unsigned char sense_L = ADCL;
						unsigned char sense_H = ADCH;
						unsigned int sense = (sense_H << 8) | sense_L;

						//	Wasser Pumpen
						//	(Die Ansteuerung ist zwar hässlich, funktioniert aber... :-))
						if (sense <= 100)
						{
							clear_bit(TRIAC_PUMP_w, TRIAC_PUMP_pin);
							_delay_ms(3);
							set_bit(TRIAC_PUMP_w, TRIAC_PUMP_pin);
						}

						//	Der Benutzer kann das Gerät mitten im Vorgang abschalten
						if (button_power_counter > BUTTON_THRESHOLD) escape = TRUE;
					}

					//	Pumpe aus (falls sie vorher eingeschaltet war)
					set_bit(TRIAC_PUMP_w, TRIAC_PUMP_pin);

					//	Kaffeewunsch erledigt
					make_coffee = 0;
				}
			}
			else
			//	Temperatur zu niedrig -> Heizen
			{
				//	Boiler an
				clear_bit(TRIAC_BOILER_w, TRIAC_BOILER_pin);

				led_green_blink = TRUE;
			}
		}
		else
		//	Wasserstand zu niedrig -> Nichts tun
		{
			//	Boiler aus (falls er eingeschaltet war)
			set_bit(TRIAC_BOILER_w, TRIAC_BOILER_pin);

			led_green_blink = FALSE;
			led_green_on = FALSE;
			led_red_blink = TRUE;
		}

	}
}


//	Initialisierung
void init()
{
	//	Nulldurchgangserkennung
	clear_bit(ZERO_CROSSING_ddr, ZERO_CROSSING_pin);
	clear_bit(ZERO_CROSSING_w, ZERO_CROSSING_pin);

	//	Taster, Interne Pull-Ups
	clear_bit(BUTTON_1_CUP_ddr, BUTTON_1_CUP_pin);	//	1 Tasse
	set_bit(BUTTON_1_CUP_w, BUTTON_1_CUP_pin);
	clear_bit(BUTTON_2_CUP_ddr, BUTTON_2_CUP_pin);	//	2 Tassen
	set_bit(BUTTON_2_CUP_w, BUTTON_2_CUP_pin);
	clear_bit(BUTTON_POWER_ddr, BUTTON_POWER_pin);	//	Power
	set_bit(BUTTON_POWER_w, BUTTON_POWER_pin);

	//	LEDs
	set_bit(LED_RED_ddr, LED_RED_pin);		//	Rot
	clear_bit(LED_RED_w, LED_RED_pin);
	set_bit(LED_GREEN_ddr, LED_GREEN_pin);	//	Grün
	clear_bit(LED_GREEN_w, LED_GREEN_pin);
	set_bit(LED_BLUE_ddr, LED_BLUE_pin);	//	Blau
	clear_bit(LED_BLUE_w, LED_BLUE_pin);

	//	Sensoren
	clear_bit(SENSOR_MAGNET_ddr, SENSOR_MAGNET_pin);	//	Hallschalter
	clear_bit(SENSOR_MAGNET_w, SENSOR_MAGNET_pin);
	clear_bit(SENSOR_TEMP_ddr, SENSOR_TEMP_pin);		//	NTC
	clear_bit(SENSOR_TEMP_w, SENSOR_TEMP_pin);

	//	TRIACs
	set_bit(TRIAC_BOILER_ddr, TRIAC_BOILER_pin);	//	Boiler
	set_bit(TRIAC_BOILER_w, TRIAC_BOILER_pin);
	set_bit(TRIAC_PUMP_ddr, TRIAC_PUMP_pin);		//	Pumpe
	set_bit(TRIAC_PUMP_w, TRIAC_PUMP_pin);

	//	ADC
	ADCSR = (1 << ADEN) | (1 << ADPS1);	//	ADC Enable, ADC Prescaler Division Factor 4

	//	TIMER1
	set_bit(TCCR1B, CTC1);		//	CTC-Mode
	clear_bit(TCCR1B, CS11);	//	Prescaler 8
	set_bit(TCCR1B, CS12);
	clear_bit(TCCR1B, CS11);
	clear_bit(TCCR1B, CS10);
	OCR1C = 124;				//	Periodendauer 1 ms

	cli();

	//	INT0 deaktivieren
	clear_bit(GIMSK, INT0);

	//	TIMER1 aktivieren
	set_bit(TIMSK, TOIE1);

	sei();
}


//	Abschalten (Schlafmodus)
void power_off()
{
	cli();
	set_bit(GIMSK, INT0);
	clear_bit(TIMSK, TOIE1);
	sei();

	//	LEDs aus
	clear_bit(LED_RED_w, LED_RED_pin);
	clear_bit(LED_GREEN_w, LED_GREEN_pin);
	clear_bit(LED_BLUE_w, LED_BLUE_pin);

	//	Schlafen, bis der Power-Button gedrückt wird
	set_bit(MCUCR, SM1);	//	Power-down Mode
	clear_bit(MCUCR, SM0);
	set_bit(MCUCR, SE);
	asm volatile("sleep"::);

	//	Hier geht es nach dem Aufwachen dann wieder weiter
	time_counter = 0;

	cli();

	//	INT0 deaktivieren
	clear_bit(GIMSK, INT0);

	//	TIMER1 aktivieren
	set_bit(TIMSK, TOIE1);

	sei();
}


//	Dient nur zum Aufwachen
ISR(INT0_vect)
{
}


//	Millisekundenzähler, LED-Steuerung, Tastenzähler
ISR(TIMER1_OVF1_vect)
{
	//	Globaler Millisekundenzähler
	if (time_counter < 1000)
	{
		time_counter++;
	}
	else
	{
		time_counter = 0;
	}

	//	Universeller Millisekundenzähler
	user_time_counter++;

	//	LEDs setzen und blinken lassen
	unsigned char leds_blink_on;

	if (time_counter >= 0 && time_counter < 499) leds_blink_on = TRUE;
	if (time_counter >= 499 && time_counter < 1000) leds_blink_on = FALSE;

	if ((led_red_on && !led_red_blink)		||	(led_red_blink && (leds_blink_on ^ led_red_blink_offset)))		set_bit(LED_RED_w, LED_RED_pin);
	if ((led_green_on && !led_green_blink)	||	(led_green_blink && (leds_blink_on ^ led_green_blink_offset)))	set_bit(LED_GREEN_w, LED_GREEN_pin);
	if ((led_blue_on && !led_blue_blink)	||	(led_blue_blink && (leds_blink_on ^ led_blue_blink_offset)))	set_bit(LED_BLUE_w, LED_BLUE_pin);

	if ((!led_red_on && !led_red_blink)		||	(led_red_blink && !(leds_blink_on ^ led_red_blink_offset)))		clear_bit(LED_RED_w, LED_RED_pin);
	if ((!led_green_on && !led_green_blink)	||	(led_green_blink && !(leds_blink_on ^ led_green_blink_offset)))	clear_bit(LED_GREEN_w, LED_GREEN_pin);
	if ((!led_blue_on && !led_blue_blink)	||	(led_blue_blink && !(leds_blink_on ^ led_blue_blink_offset)))	clear_bit(LED_BLUE_w, LED_BLUE_pin);

	//	Tastenzähler

	//	Taster "1 Tasse"
	if (bit_is_clear(BUTTON_1_CUP_r, BUTTON_1_CUP_pin))
	{ if (button_1_cup_counter < 65535) button_1_cup_counter++; }
	else
	{ if (button_1_cup_counter > 0) button_1_cup_counter--; }

	//	Taster "2 Tassen"
	if (bit_is_clear(BUTTON_2_CUP_r, BUTTON_2_CUP_pin))
	{ if (button_2_cup_counter < 65535) button_2_cup_counter++; }
	else
	{ if (button_2_cup_counter > 0) button_2_cup_counter--; }

	//	Taster "Power"
	if (bit_is_clear(BUTTON_POWER_r, BUTTON_POWER_pin))
	{ if (button_power_counter < 65535) button_power_counter++; }
	else
	{ if (button_power_counter > 0) button_power_counter--; }
}
