/*
    Copyright (C) 2012 RedoX <dev@redox.ws>
    This file is part of SenseoMod.

        SenseoMod is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        (at your option) any later version.

        SenseoMod is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with SenseoMod.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>

#include "main.h"

volatile unsigned char InputSwitches,MakeCoffee,Leds,LedsCounter;
volatile uint16_t TimerLimit;

int __attribute__((noreturn)) main(void) {
	wdt_enable(WDTO_1S);
	uint16_t sensor;
	unsigned char SenseOk;
																	   /* WGM = 0100, Mode 4, CTC */
	GTCCR |= _BV(TSM);															 /* Disable Timer */
	TCCR1B = _BV(WGM12) | _BV(CS11);				 		 /* TimerPrescaler = 8 -> Freq = 1MHz */
	OCR1AH = TimerInterval>>8; OCR1AL = TimerInterval&0xFF;	 				 /*  Defines INT Freq */
	GTCCR &= ~_BV(TSM);															  /* Enable Timer */
	GIMSK = _BV(PCIE0);													 	 /* Interrupts PCINT0 */
																			/* Pins Configuration */
	LedsDir = LedsMask; LedsPort = 0;										 /* PortB, Output low */
	InpDir = 0; InpPort = InpMask;									/* PortA, Inputs (except ADC) */
	CoffeeDir |= PumpIO | BoilerIO; CoffeePort |= PumpIO|BoilerIO;			 /* PortA, Output low */
	DIDR0 = TempADC|WaterLvlADC;										 /* Disable Digital Input */
	InputSwitches = PowerSwIO;										   			  /* Start Asleep */
	MakeCoffee = SenseOk = 0x00;
	sei();
	
	while(1) {
		wdt_reset();														  /* Keep Running boy */
		
			/***************************************************************************** Inputs */
		if(InputSwitches & PowerSwIO) {
			LedsPort=LedsMask; _delay_ms(500); LedsPort&=~LedsMask;				         /* Blink */
			if(!(MakeCoffee&StOffMode)) {
				CoffeePort |= (PumpIO|BoilerIO);						/* Turn Off Pump & Boiler */
				PCMSK0 = PowerSwIO|AutoIO; 								  /* Discard Cups buttons */
				TIMSK1 &= ~_BV(OCIE1A);								   /* Discard Timer Interrupt */
				MakeCoffee = 0x00;
			} else {
				if(InputSwitches & TwoSwIO) {
					CoffeePort &= ~PumpIO;
					while(getADCValue(WaterLvlADC) > WLvl_MinVal && (~InpPin & TwoSwIO))
						wdt_reset();
					CoffeePort |= PumpIO;
				}
				PCMSK0 = InpMask;					  /* Enable the Interrupts for the Input Pins */
				TIMSK1 |= _BV(OCIE1A);					/* Interrupt when the Timer matches OCR1A */
				Leds = 0; LedsCounter = 0; TimerLimit = 0;		   				   /* Reset Stuff */
			}
			MakeCoffee ^= StOffMode;
			InputSwitches &= ~PowerSwIO;								/* Switch input processed */
		} else if(InputSwitches & AutoIO) {
			MakeCoffee |= StAuto;										/* Can't be auto-disabled */
			if(MakeCoffee&StOffMode) 								  /* Asleep ? Have to Wake up */
				InputSwitches |= PowerSwIO;
			InputSwitches &= ~AutoIO;										   /* Input processed */
		}
		
		if(InputSwitches & OneSwIO) {
			MakeCoffee ^= StOneCup;												  /* Fill One cup */
			InputSwitches &= ~OneSwIO;									/* Switch input processed */
		} else if(InputSwitches & TwoSwIO) {
			MakeCoffee ^= StTwoCups;											 /* Fill Two Cups */
			InputSwitches &= ~TwoSwIO;									/* Switch input processed */
		} 
		  /***************************************************************************** Action ! */
		if(!(MakeCoffee & StOffMode)) {
			sensor = getADCValue(TempADC);								   /* Get the Temperature */
			if(sensor > Temp_MinVal) {
														  /* Just when here ? Stabilization delay */
				if(!(SenseOk & TempIO) && !(MakeCoffee & StTempWait) && !(SenseOk & TempWait)) {
					TimerLimit = TempSetSecValue;
					SenseOk |= TempWait;
					MakeCoffee |= StTempWait;
				} else if(!(SenseOk & TempIO) && (SenseOk & TempWait) && !(MakeCoffee & StTempWait)) {
					SenseOk |= TempIO;										 /* Temperature is Ok */
				}
			} else {
				MakeCoffee &= ~StTempWait;
				SenseOk &= ~(TempIO|TempWait);							 /* Temperature is not Ok */
			}
			
			if(sensor > Temp_MaxVal)
				CoffeePort |= BoilerIO;											 /* Too hot man ! */

			sensor = getADCValue(WaterLvlADC);								 /* Get the Water Lvl */
			if(sensor >= WLvl_MinVal) {
				if(SenseOk & TempIO) {
					if(MakeCoffee & StCoffeePaused) {								  /* Paused ? */
						MakeCoffee ^= StCoffeePaused;								   /* Restart */
						GTCCR &= ~_BV(TSM);							   /* Start the Timer/Counter */
						CoffeePort &= ~PumpIO;							 /* Keep making Coffee :) */
					} 
					if(MakeCoffee & (StOneCup|StTwoCups|StAuto)) {				 /* Make Coffee ? */
						if(!(MakeCoffee & StCoffeeStarted)) {					  /* First Time ? */
							MakeCoffee &= ~StIdleToSleep;			/* If was idle, disable delay */
							if(MakeCoffee & StOneCup)
								TimerLimit = OneValue;							/* Go for One Cup */
							else if(MakeCoffee & StTwoCups)
								TimerLimit = TwoValue;								 /* Or two ^^ */
							else if(MakeCoffee & StAuto)
								TimerLimit = AutoValue;								/* AutoChoice */
							CoffeePort &= ~PumpIO;					   /* Go Shadok, Go ! Pump On */
							MakeCoffee |= StCoffeeStarted;					  /* We've been there */
						}
						setLeds(MakeCoffee & StAuto ? LedG : LBlinkB);		/* Led Coffee Running */
					} else {
						setLeds(LedB);										    /* Led Idle Ready */
						if(!(MakeCoffee & StIdleToSleep)) {
							TimerLimit = SleepValue;						/* Delay before sleep */
							MakeCoffee |= StIdleToSleep;						 	 /* Wait idle */
							MakeCoffee &= ~StCoffeeStarted;					
						}
						CoffeePort |= PumpIO;
					}
				} else {
					setLeds(LBlinkG);										  /* Led blinks Green */
					if((MakeCoffee & StCoffeeStarted) && 
					   !(MakeCoffee & StCoffeePaused))  		 /* Coffee Running ? Keep pumping */
						CoffeePort &= ~PumpIO;									  /* Stop Pumping */
					else
						CoffeePort |= PumpIO;									  /* Stop Pumping */
					CoffeePort &= ~BoilerIO;							   /* We need Hot water ! */
				}
				SenseOk |= WaterLvlIO;										/* Still enough water */
			} else {
				setLeds(LBlinkR);												/* Led blinks Red */
				CoffeePort |= (BoilerIO|PumpIO);							/* Stop Boiler & Pump */
				if( !(MakeCoffee & StCoffeePaused) && 
					(MakeCoffee & StCoffeeStarted)) {			  			  /* Coffee Running ? */
					MakeCoffee |= StCoffeePaused;									  /* Pause It */
					GTCCR |= _BV(TSM);								   	/* Stop the Timer/Counter */
				}
				SenseOk &= ~WaterLvlIO;										  /* Not enough water */
			}
		}
	}
}
		/********************************** Interrupt Handler for PCINT0:7, we only need PCINT4:7 */
ISR(PCINT0_vect) {
	InputSwitches = (~InpPin)&InpMask;		  /* Invert InpPin since we use the internal Pull-Ups */
	_delay_ms(250);

	if(((~InpPin)&InpMask) != InputSwitches) {
		InputSwitches = 0;
	}
}
		/***************************************** Interrupt Handler for TIMER1 used as a counter */
ISR(TIM1_COMPA_vect) {
	if(++LedsCounter > LBlinkLim) {									  /* Handles the LED Blinking */
		LedsCounter = 0;													 /* Reset the Counter */
		if((Leds>>4) & LedsMask)
			LedsPort ^= (Leds>>4) & LedsMask;										 	 /* Blink */
		else
			LedsPort = Leds & LedsMask;													/* Steady */
	}
	if(TimerLimit)
		TimerLimit--;

	if((MakeCoffee & StTempWait) && (TimerLimit < 1))
		MakeCoffee &= ~StTempWait;
	
	if((MakeCoffee & StIdleToSleep) && (TimerLimit < 1)) {
		MakeCoffee &= ~StIdleToSleep;
		InputSwitches = PowerSwIO;										 /* Delay has passed away */
	}

	if((MakeCoffee & StCoffeeStarted) && (TimerLimit < 1)) {		 /* Handles stopping the Pump */
		if(MakeCoffee&StAuto)														/* AutoMode ? */
			InputSwitches = PowerSwIO;											  /* Shutdown now */
		MakeCoffee &= ~(StOneCup|StTwoCups|StAuto|StCoffeeStarted);
	}

}

		/********************************************************************** Helpers fundtions */
uint16_t getADCValue(unsigned char io) {
	uint16_t r=0;
	ADMUX = io;																   /* Use Vcc as Vref */
	ADCSRA = _BV(ADEN) | _BV(ADSC);							/* Default settings, start Conversion */
	while(ADCSRA&_BV(ADSC));											 /* Wait until completion */
	r = ADCL | (ADCH<<8);														/* Get the result */
	ADCSRA &= ~_BV(ADEN);													   /* Disable the ADC */
	return r;
}

void setLeds(unsigned char v) {
	if(Leds != v) {														 /* If there's any change */
		LedsPort &= ~Leds;														/* Clear the Port */
		Leds = v;															  /* Update the Value */
	}
}
