/*
    Copyright (C) 2012 RedoX <dev@redox.ws>
    This file is part of SenseoMod.

        SenseoMod is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        (at your option) any later version.

        SenseoMod is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with SenseoMod.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _Main_h
#define _Main_h

/**************************************************************************************************/
/********************************************* Config *********************************************/
/**************************************************************************************************/
#define WLvl_MinVal			470					/* Min Value for the Hall Sensor of the Water Lvl */
#define Temp_MinVal			386 			/* Min Value returned by the ADC, for the Temperature */
#define Temp_MaxVal			420								 	 /* Maximum value, security stuff */
#define LBlinkLim			25		  /* Led Update Interval, in INT count = 25 -> (Tled = 500ms) */
#define	SleepSec			120									  /* Delay before sleep when idle */
#define	TempSetSec		15

#define OneSec				16											   /* 18s to fill one Cup */
#define TwoSec				(2*OneSec)							   		   		   /* 33s for two */
#define AutoSec				TwoSec											  /* Auto => Two Cups */

/**************************************************************************************************/
/****************************************** PreComputing ******************************************/
/**************************************************************************************************/
								   /* Value = (ValueSec*1000)/(TimerInterval/1000) => ValueSec*50 */
#define Coeff				50
#define OneValue			(OneSec * Coeff)
#define TwoValue			(TwoSec * Coeff)
#define AutoValue			(AutoSec * Coeff)

#define SleepValue			(SleepSec * Coeff)
#define TempSetSecValue			(TempSetSec * Coeff)

/**************************************************************************************************/
/*********************************************** IOs **********************************************/
/**************************************************************************************************/

#define LedsDir				DDRB
#define LedsPort			PORTB
#define LedsMask			0b00000111
#define LedR				0x4															   /* PB2 */
#define LedG				0x2															   /* PB1 */
#define LedB				0x1															   /* PB0 */

#define InpDir				DDRA
#define InpPort				PORTA
#define InpPin				PINA
#define	InpMask				0b11110000
#define PowerSwIO			0x20														   /* PA5 */
#define OneSwIO				0x80														   /* PA7 */
#define TwoSwIO				0x10														   /* PA4 */
#define AutoIO				0x40														   /* PA6 */

#define CoffeeDir			DDRA
#define CoffeePin			PINA
#define CoffeePort			PORTA
#define CoffeeMask			0b00001111
#define WaterLvlIO			0x1															   /* PA0 */
#define WaterLvlADC			0x0															  /* ADC0 */
#define PumpIO				0x2															   /* PA1 */
#define TempIO				0x4															   /* PA2 */
#define TempADC				0x2															  /* ADC2 */
#define BoilerIO			0x8															   /* PA3 */

/**************************************************************************************************/
/******************************************** Constants *******************************************/
/**************************************************************************************************/
									 /* These three are defined according to the Leds pins values */
#define LBlinkR				0x40
#define LBlinkG				0x20
#define LBlinkB				0x10

#define TempWait			0x40

/* MakeCoffee 
 * 0x01: OneCup
 * 0x02: TwoCups
 * 0x04: AutoCup
 * 0x08: CoffeeHasStarted
 * 0x10: CoffeePaused
 * 0x20: TempWait
 * 0x40: Idle to Sleep
 * 0x80: Off mode (Boiler, Pump, Leds)
 */
#define StOneCup 			0x01
#define StTwoCups			0x02
#define StAuto				0x04
#define StCoffeeStarted		0x08
#define StCoffeePaused		0x10
#define StTempWait 			0x20
#define StIdleToSleep		0x40
#define StOffMode			0x80

#define TimerInterval		20000  /* Timer runs at 1MHz, so we get an INT Freq of 50 Hz (T=20ms) */

/**************************************************************************************************/
/******************************************* Prototypes *******************************************/
/**************************************************************************************************/

void WakeUp(void);
void Shutdown(void);
uint16_t getADCValue(unsigned char io);
void setLeds(unsigned char v);

#endif